@@ -15,14 +15,16 @@ import httplib
 from datetime import datetime
 from collections import namedtuple
 from xml.etree import cElementTree as ET
+from xml.etree.ElementTree import ParseError
 from twisted.internet import reactor, defer
 from twisted.internet.protocol import Protocol
 from twisted.web.client import Agent
 from twisted.internet.ssl import ClientContextFactory
 from twisted.web.http_headers import Headers
+from twisted.internet.threads import deferToThread
 from . import constants as c
 
-from .krb5 import kinit, ccname
+from .krb5 import kinit, ccname, add_trusted_realm
 
 # ZEN-15434 lazy import to avoid segmentation fault during install
 kerberos = None
@@ -44,14 +46,23 @@ _REQUEST_TEMPLATES = {}
 _CONTENT_TYPE = {'Content-Type': ['application/soap+xml;charset=UTF-8']}
 _MAX_KERBEROS_RETRIES = 3
 _MARKER = object()
-_ENCRYPTED_CONTENT_TYPE={"Content-Type" : ["multipart/encrypted;protocol=\"application/HTTP-Kerberos-session-encrypted\";boundary=\"Encrypted Boundary\""]}
-_BODY="""--Encrypted Boundary
+_ENCRYPTED_CONTENT_TYPE = {
+    "Content-Type": [
+        "multipart/encrypted;"
+        "protocol=\"application/HTTP-Kerberos-session-encrypted\";"
+        "boundary=\"Encrypted Boundary\""]}
+_BODY = """--Encrypted Boundary
 Content-Type: application/HTTP-Kerberos-session-encrypted
 OriginalContent: type=application/soap+xml;charset=UTF-8;Length={original_length}
 --Encrypted Boundary
 Content-Type: application/octet-stream
 {emsg}--Encrypted Boundary
 """
+
+_KRB_INTERNAL_CACHE_ERR = 'Internal credentials cache error while storing '\
+    'credentials while getting initial credentials'
+
+
 def _has_get_attr(obj, attr_name):
     attr_value = getattr(obj, attr_name, _MARKER)
     if attr_value is _MARKER:
@@ -67,6 +78,7 @@ class MyWebClientContextFactory(object):
     def getContext(self, hostname, port):
         return self._options.getContext()
 
+
 def _get_agent():
     context_factory = MyWebClientContextFactory()
     try:
@@ -76,12 +88,13 @@ def _get_agent():
         pool.maxPersistentPerHost = _MAX_PERSISTENT_PER_HOST
         pool.cachedConnectionTimeout = _CACHED_CONNECTION_TIMEOUT
         agent = Agent(reactor, context_factory,
-                       connectTimeout=_CONNECT_TIMEOUT, pool=pool)
+                      connectTimeout=_CONNECT_TIMEOUT, pool=pool)
     except ImportError:
         from _zenclient import ZenAgent
         agent = ZenAgent(reactor, context_factory, persistent=True, maxConnectionsPerHostName=1)
     return agent
 
+
 class _StringProducer(object):
     """
     The length attribute must be a non-negative integer or the constant
@@ -109,12 +122,16 @@ class _StringProducer(object):
     def stopProducing(self):
         pass
 
+
 def _parse_error_message(xml_str):
     if not xml_str:
         return ""
-    elem = ET.fromstring(xml_str)
-    text = elem.findtext('.//{' + c.XML_NS_SOAP_1_2 + '}Text').strip()
-    detail = elem.findtext('.//{' + c.XML_NS_SOAP_1_2 + '}Detail/*/*').strip()
+    try:
+        elem = ET.fromstring(xml_str)
+        text = elem.findtext('.//{' + c.XML_NS_SOAP_1_2 + '}Text').strip()
+        detail = elem.findtext('.//{' + c.XML_NS_SOAP_1_2 + '}Detail/*/*').strip()
+    except ParseError:
+        return "Malformed XML: {}".format(xml_str)
     return "{0} {1}".format(text, detail)
 
 
@@ -177,7 +194,7 @@ class AuthGSSClient(object):
         1  : GSSAPI step complete, or function return OK
     """
 
-    def __init__(self, service, username, password, dcip):
+    def __init__(self, service, conn_info):
         """
         @param service: a string containing the service principal in the form
             'type@fqdn' (e.g. 'imap@mail.apple.com').
@@ -188,20 +205,28 @@ class AuthGSSClient(object):
         if not kerberos:
             import kerberos
         self._service = service
-        self._username = username
-        self._password = password
-        self._dcip = dcip
-        gssflags = kerberos.GSS_C_CONF_FLAG|kerberos.GSS_C_MUTUAL_FLAG|kerberos.GSS_C_SEQUENCE_FLAG|kerberos.GSS_C_INTEG_FLAG
+        self._conn_info = conn_info
+        self._username = conn_info.username
+        self._password = conn_info.password
+        self._dcip = conn_info.dcip
+        gssflags = kerberos.GSS_C_CONF_FLAG | kerberos.GSS_C_MUTUAL_FLAG | kerberos.GSS_C_SEQUENCE_FLAG | kerberos.GSS_C_INTEG_FLAG
 
-        os.environ['KRB5CCNAME'] = ccname(username)
-        result_code, self._context = kerberos.authGSSClientInit(service,gssflags=gssflags)
+        os.environ['KRB5CCNAME'] = ccname(conn_info.username)
+        if conn_info.trusted_realm and conn_info.trusted_kdc:
+            add_trusted_realm(conn_info.trusted_realm, conn_info.trusted_kdc)
+        if hasattr(kerberos, 'authGSSClientWrapIov'):
+            result_code, self._context = kerberos.authGSSClientInit(service, gssflags=gssflags)
+        else:
+            result_code, self._context = kerberos.authGSSClientInit(service)
         if result_code != kerberos.AUTH_GSS_COMPLETE:
             raise Exception('kerberos authGSSClientInit failed')
 
     def __del__(self):
-        result_code = kerberos.authGSSClientClean(self._context)
-        if result_code != kerberos.AUTH_GSS_COMPLETE:
-            raise Exception('kerberos authGSSClientClean failed')
+        if self._context is not None:
+            result_code = kerberos.authGSSClientClean(self._context)
+
+            if result_code != kerberos.AUTH_GSS_COMPLETE:
+                raise Exception('kerberos authGSSClientClean failed')
 
     def _step(self, challenge=''):
         """
@@ -213,7 +238,7 @@ class AuthGSSClient(object):
         @return:          a result code
         """
         log.debug('GSSAPI step challenge="{0}"'.format(challenge))
-        return kerberos.authGSSClientStep(self._context, challenge)
+        return deferToThread(kerberos.authGSSClientStep, self._context, challenge)
 
     @defer.inlineCallbacks
     def get_base64_client_data(self, challenge=''):
@@ -224,15 +249,21 @@ class AuthGSSClient(object):
         result_code = None
         for i in xrange(_MAX_KERBEROS_RETRIES):
             try:
-                result_code = self._step(challenge)
+                result_code = yield self._step(challenge)
                 break
             except kerberos.GSSError as e:
                 msg = e.args[1][0]
-                if msg == 'Cannot determine realm for numeric host address' or \
-                   msg == 'Server not found in Kerberos database':
-                       raise Exception(msg)
+                if msg == 'Cannot determine realm for numeric host address':
+                    raise Exception(msg)
+                elif msg == 'Server not found in Kerberos database':
+                    raise Exception(msg + ': ' + self._service)
                 log.debug('{0}. Calling kinit.'.format(msg))
-                yield kinit(self._username, self._password, self._dcip)
+                kinit_result = yield kinit(self._username, self._password, self._dcip)
+                if kinit_result:
+                    # this error is ok.  it just means more
+                    # than one process is calling kinit
+                    if _KRB_INTERNAL_CACHE_ERR not in kinit_result:
+                        raise Exception(kinit_result)
 
         if result_code != kerberos.AUTH_GSS_CONTINUE:
             raise Exception('kerberos authGSSClientStep failed ({0}).'
@@ -240,6 +271,7 @@ class AuthGSSClient(object):
         base64_client_data = kerberos.authGSSClientResponse(self._context)
         defer.returnValue(base64_client_data)
 
+    @defer.inlineCallbacks
     def get_username(self, challenge):
         """
         Get the user name of the principal authenticated via the now complete
@@ -248,12 +280,12 @@ class AuthGSSClient(object):
         @param challenge: a string containing the base64-encoded server data
         @return:          a string containing the user name.
         """
-        result_code = self._step(challenge)
+        result_code = yield self._step(challenge)
         if result_code != kerberos.AUTH_GSS_COMPLETE:
             raise Exception('kerberos authGSSClientStep failed ({0}). '
                             'challenge={1}'
                             .format(result_code, challenge))
-        return kerberos.authGSSClientUserName(self._context)
+        defer.returnValue(kerberos.authGSSClientUserName(self._context))
 
     def encrypt_body(self, body):
         # get original length of body. wrap will encrypt in place
@@ -262,34 +294,41 @@ class AuthGSSClient(object):
         ebody = base64.b64encode(body)
         # wrap it up
         try:
-            rc,pad_len = kerberos.authGSSClientWrapIov(self._context,ebody,1)
+            rc, pad_len = kerberos.authGSSClientWrapIov(self._context, ebody, 1)
             if rc is not kerberos.AUTH_GSS_COMPLETE:
                 log.debug("Unable to encrypt message body")
                 return
         except AttributeError:
             # must be on centos 5, encryption not possible
             return body
+        except kerberos.GSSError as e:
+            msg = e.args[1][0]
+            raise Exception(msg)
 
         # get wrapped request which is in b64 encoding
         ewrap = kerberos.authGSSClientResponse(self._context)
         # decode wrapped request
         payload = bytes(base64.b64decode(ewrap))
         # add carriage returns to body
-        body = _BODY.replace('\n','\r\n')
-        body = bytes(body.format(original_length=orig_len+pad_len,emsg=payload))
+        body = _BODY.replace('\n', '\r\n')
+        body = bytes(body.format(original_length=orig_len + pad_len, emsg=payload))
         return body
 
     def decrypt_body(self, body):
         try:
             b_start = body.index("Content-Type: application/octet-stream") + \
-                      len("Content-Type: application/octet-stream\r\n")
+                len("Content-Type: application/octet-stream\r\n")
         except ValueError:
             # Unencrypted data, return body
             return body
-        b_end = body.index("--Encrypted Boundary",b_start)
+        b_end = body.index("--Encrypted Boundary", b_start)
         ebody = body[b_start:b_end]
         ebody = base64.b64encode(ebody)
-        rc = kerberos.authGSSClientUnwrapIov(self._context, ebody)
+        try:
+            rc = kerberos.authGSSClientUnwrapIov(self._context, ebody)
+        except kerberos.GSSError as e:
+            msg = e.args[1][0]
+            raise Exception(msg)
         if rc is not kerberos.AUTH_GSS_COMPLETE:
             log.debug("Unable to decrypt message body")
             return
@@ -297,15 +336,31 @@ class AuthGSSClient(object):
         body = base64.b64decode(ewrap)
         return body
 
+    def cleanup(self):
+        kerberos.authGSSClientClean(self._context)
+        self._context = None
+
+
+def get_auth_details(auth_header=''):
+    auth_details = ''
+    for field in auth_header.split(','):
+        try:
+            kind, details = field.strip().split(' ', 1)
+            if kind.lower() == 'kerberos':
+                auth_details = details.strip()
+                break
+        except ValueError:
+            continue
+    return auth_details
+
+
 @defer.inlineCallbacks
 def _authenticate_with_kerberos(conn_info, url, agent, gss_client=None):
-    service = '{0}@{1}'.format(conn_info.scheme.upper(), conn_info.hostname)
+    service = '{0}@{1}'.format(conn_info.service.upper(), conn_info.hostname)
     if gss_client is None:
         gss_client = AuthGSSClient(
             service,
-            conn_info.username,
-            conn_info.password,
-            conn_info.dcip)
+            conn_info)
 
     base64_client_data = yield gss_client.get_base64_client_data()
     auth = 'Kerberos {0}'.format(base64_client_data)
@@ -313,9 +368,19 @@ def _authenticate_with_kerberos(conn_info, url, agent, gss_client=None):
     k_headers.addRawHeader('Authorization', auth)
     k_headers.addRawHeader('Content-Length', '0')
     response = yield agent.request('POST', url, k_headers, None)
+    auth_header = response.headers.getRawHeaders('WWW-Authenticate')[0]
+    auth_details = get_auth_details(auth_header)
+
     if response.code == httplib.UNAUTHORIZED:
+        try:
+            if auth_details:
+                yield gss_client._step(auth_details)
+        except kerberos.GSSError as e:
+            msg = "HTTP Unauthorized received on kerberos initialization.  "\
+                "Kerberos error code {0}: {1}.".format(e.args[1][1], e.args[1][0])
+            raise Exception(msg)
         raise UnauthorizedError(
-            "HTTP Unauthorized received on initial kerberos request.")
+            "HTTP Unauthorized received on initial kerberos request.  Check username and password")
     elif response.code == httplib.FORBIDDEN:
         raise ForbiddenError(
             "Forbidden. Check WinRM port and version.")
@@ -323,26 +388,21 @@ def _authenticate_with_kerberos(conn_info, url, agent, gss_client=None):
         proto = _StringProtocol()
         response.deliverBody(proto)
         xml_str = yield proto.d
-        xml_str = gss_client.decrypt_body(gss_client._context, xml_str)
+        xml_str = gss_client.decrypt_body(xml_str)
         raise Exception(
             "status code {0} received on initial kerberos request {1}"
             .format(response.code, xml_str))
-    auth_header = response.headers.getRawHeaders('WWW-Authenticate')[0]
-    for field in auth_header.split(','):
-        kind, details = field.strip().split(' ', 1)
-        if kind.lower() == 'kerberos':
-            auth_details = details.strip()
-            break
-    else:
+    if not auth_details:
         raise Exception(
             'negotiate not found in WWW-Authenticate header: {0}'
             .format(auth_header))
-    k_username = gss_client.get_username(auth_details)
+    k_username = yield gss_client.get_username(auth_details)
     log.debug('kerberos auth successful for user: {0} / {1} '
               .format(conn_info.username, k_username))
     defer.returnValue(gss_client)
 
-ConnectionInfo = namedtuple(
+
+class ConnectionInfo(namedtuple(
     'ConnectionInfo', [
         'hostname',
         'auth_type',
@@ -352,7 +412,23 @@ ConnectionInfo = namedtuple(
         'port',
         'connectiontype',
         'keytab',
-        'dcip'])
+        'dcip',
+        'timeout',
+        'trusted_realm',
+        'trusted_kdc',
+        'ipaddress',
+        'service'])):
+    def __new__(cls, hostname, auth_type, username, password, scheme, port,
+                connectiontype, keytab, dcip, timeout=60, trusted_realm='', trusted_kdc='', ipaddress='', service=''):
+        if not ipaddress:
+            ipaddress = hostname
+        if not service:
+            service = scheme
+        return super(ConnectionInfo, cls).__new__(cls, hostname, auth_type,
+                                                  username, password, scheme,
+                                                  port, connectiontype, keytab,
+                                                  dcip, timeout,
+                                                  trusted_realm, trusted_kdc, ipaddress, service)
 
 
 def verify_hostname(conn_info):
@@ -361,6 +437,12 @@ def verify_hostname(conn_info):
         raise Exception("hostname is not resolvable")
 
 
+def verify_ipaddress(conn_info):
+    has_ipaddress, ipaddress = _has_get_attr(conn_info, 'ipaddress')
+    if not has_ipaddress or not ipaddress:
+        raise Exception("ipaddress missing")
+
+
 def verify_auth_type(conn_info):
     has_auth_type, auth_type = _has_get_attr(conn_info, 'auth_type')
     if not has_auth_type or auth_type not in ('basic', 'kerberos'):
@@ -388,6 +470,17 @@ def verify_scheme(conn_info):
             .format(scheme))
 
 
+def verify_service(conn_info):
+    has_service, service = _has_get_attr(conn_info, 'service')
+    if not has_service:
+        # if not supplied, default to scheme
+        has_service, service = _has_get_attr(conn_info, 'scheme')
+    if not has_service or service not in ['http', 'https', 'wsman']:
+        raise Exception(
+            "service must be http, https, or wsman: {0}"
+            .format(service))
+
+
 def verify_port(conn_info):
     has_port, port = _has_get_attr(conn_info, 'port')
     if not has_port or not port or not isinstance(port, int):
@@ -400,14 +493,24 @@ def verify_connectiontype(conn_info):
         raise Exception("connectiontype missing")
 
 
+def verify_timeout(conn_info):
+    has_timeout, timeout = _has_get_attr(conn_info, 'timeout')
+    if not has_timeout:
+        raise Exception("timeout missing")
+    if not timeout:
+        conn_info.timeout = 60
+
+
 def verify_conn_info(conn_info):
     verify_hostname(conn_info)
+    verify_ipaddress(conn_info)
     verify_auth_type(conn_info)
     verify_username(conn_info)
     verify_password(conn_info)
     verify_scheme(conn_info)
     verify_port(conn_info)
     verify_connectiontype(conn_info)
+    verify_timeout(conn_info)
 
 
 class RequestSender(object):
@@ -423,7 +526,7 @@ class RequestSender(object):
 
     @defer.inlineCallbacks
     def _get_url_and_headers(self):
-        url = "{c.scheme}://{c.hostname}:{c.port}/wsman".format(c=self._conn_info)
+        url = "{c.scheme}://{c.ipaddress}:{c.port}/wsman".format(c=self._conn_info)
         if self._conn_info.auth_type == 'basic':
             headers = Headers(_CONTENT_TYPE)
             headers.addRawHeader('Connection', self._conn_info.connectiontype)
@@ -464,8 +567,7 @@ class RequestSender(object):
         if self.is_kerberos():
             encrypted_request = self.gssclient.encrypt_body(request)
             if not encrypted_request.startswith("--Encrypted Boundary"):
-                self.headers = Headers(_CONTENT_TYPE)
-                self.headers.addRawHeader('Connection', self._conn_info.connectiontype)
+                self._headers.setRawHeaders('Content-Type', _CONTENT_TYPE['Content-Type'])
             body_producer = _StringProducer(encrypted_request)
         else:
             body_producer = _StringProducer(request)
@@ -476,19 +578,36 @@ class RequestSender(object):
             raise e
         log.debug('received response {0} {1}'.format(
             response.code, request_template_name))
-        if response.code == httplib.UNAUTHORIZED:
-            # check to see if we need to re-authorize due to lost connection
+        if response.code == httplib.UNAUTHORIZED or response.code == httplib.BAD_REQUEST:
+            # check to see if we need to re-authorize due to lost connection or bad request error
             if self.gssclient is not None:
                 self.agent = _get_agent()
-                yield _authenticate_with_kerberos(self._conn_info, self._url, self.agent, self.gssclient)
+                # do some cleanup first.  memory leaks were occurring
+                self.gssclient.cleanup()
+                self.gssclient = None
                 try:
+                    yield self._set_url_and_headers()
+                    encrypted_request = self.gssclient.encrypt_body(request)
+                    if not encrypted_request.startswith("--Encrypted Boundary"):
+                        self._headers.setRawHeaders('Content-Type', _CONTENT_TYPE['Content-Type'])
+                    body_producer = _StringProducer(encrypted_request)
                     response = yield self.agent.request(
                         'POST', self._url, self._headers, body_producer)
                 except Exception as e:
                     raise e
             if response.code == httplib.UNAUTHORIZED:
+                if self.is_kerberos():
+                    auth_header = response.headers.getRawHeaders('WWW-Authenticate')[0]
+                    auth_details = get_auth_details(auth_header)
+                    try:
+                        if auth_details:
+                            yield self.gssclient._step(auth_details)
+                    except kerberos.GSSError as e:
+                        msg ="HTTP Unauthorized received.  "\
+                        "Kerberos error code {0}: {1}.".format(e.args[1][1],e.args[1][0])
+                        raise Exception(msg)
                 raise UnauthorizedError(
-                    "Unauthorized: Check username and password")
+                    "HTTP Unauthorized received: Check username and password")
         if response.code == httplib.FORBIDDEN:
             raise ForbiddenError(
                 "Forbidden: Check WinRM port and version")
@@ -503,6 +622,19 @@ class RequestSender(object):
                 response.code, message))
         defer.returnValue(response)
 
+    def close_connections(self):
+        # close connections
+        # return a Deferred()
+        if self.agent and hasattr(self.agent, 'closeCachedConnections'):
+            # twisted 11 has no return and is part of the Agent
+            return defer.succeed(self.agent.closeCachedConnections())
+        elif self.agent:
+            # twisted 12 returns a Deferred
+            return self.agent._pool.closeCachedConnections()
+        else:
+            # no agent
+            return defer.succeed(None)
+
 
 class _StringProtocol(Protocol):
 
@@ -543,6 +675,11 @@ class EtreeRequestSender(object):
                 log.debug('Could not prettify response XML: "{0}"'.format(xml_str))
         defer.returnValue(ET.fromstring(xml_str))
 
+    @defer.inlineCallbacks
+    def close_connections(self):
+        closed = yield self._sender.close_connections()
+        defer.returnValue(closed)
+
 
 def create_etree_request_sender(conn_info):
     sender = RequestSender(conn_info)
